/*
 * Copyright (c) 2010-2024. Institut Pasteur.
 *
 * This file is part of Icy.
 * Icy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Icy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Icy. If not, see <https://www.gnu.org/licenses/>.
 */

package plugins.nchenouard.trackprocessorintensityprofile;

import org.bioimageanalysis.icy.model.sequence.Sequence;
import org.jfree.data.xy.XYSeries;
import plugins.fab.trackmanager.TrackSegment;
import plugins.nchenouard.spot.Detection;

import java.util.ArrayList;
import java.util.Arrays;

public class TrackAnalysisDiskAveragingNoBackground extends TrackAnalysis {
    XYSeries[] meanIntensity;
    XYSeries[] minIntensity;
    XYSeries[] maxIntensity;
    XYSeries[] medianIntensity;
    XYSeries[] sumIntensity;
    XYSeries[] varIntensity;

    TrackAnalysisDiskAveragingNoBackground(
            final TrackSegment track,
            final Sequence sequence,
            final String description,
            final AveragingType averagingMethod
    ) {
        super(track, sequence, description, averagingMethod);

        meanIntensity = new XYSeries[sequence.getSizeC()];
        for (int c = 0; c < sequence.getSizeC(); c++)
            meanIntensity[c] = new XYSeries(description);
        minIntensity = new XYSeries[sequence.getSizeC()];
        for (int c = 0; c < sequence.getSizeC(); c++)
            minIntensity[c] = new XYSeries(description);
        maxIntensity = new XYSeries[sequence.getSizeC()];
        for (int c = 0; c < sequence.getSizeC(); c++)
            maxIntensity[c] = new XYSeries(description);
        medianIntensity = new XYSeries[sequence.getSizeC()];
        for (int c = 0; c < sequence.getSizeC(); c++)
            medianIntensity[c] = new XYSeries(description);
        sumIntensity = new XYSeries[sequence.getSizeC()];
        for (int c = 0; c < sequence.getSizeC(); c++)
            sumIntensity[c] = new XYSeries(description);
        varIntensity = new XYSeries[sequence.getSizeC()];
        for (int c = 0; c < sequence.getSizeC(); c++)
            varIntensity[c] = new XYSeries(description);
    }

    @Override
    public void clearSeries() {
        for (final XYSeries xySeries : meanIntensity) xySeries.clear();
        for (final XYSeries xySeries : minIntensity) xySeries.clear();
        for (final XYSeries xySeries : maxIntensity) xySeries.clear();
        for (final XYSeries xySeries : medianIntensity) xySeries.clear();
        for (final XYSeries xySeries : sumIntensity) xySeries.clear();
        for (final XYSeries xySeries : varIntensity) xySeries.clear();
    }

    //TODO: make it 3D
    public void fillAveragingSeriesAtT(final int t, final double[][][] imageTab, final double diskRadius, final double outterDiskRadius) {
        if (track.getFirstDetection().getT() <= t && track.getLastDetection().getT() >= t) {
            final Detection d = track.getDetectionAtTime(t);
            if (d != null) {
                final int width = sequence.getSizeX();
                final int height = sequence.getSizeY();
                final int depth = sequence.getSizeZ();

                final ArrayList<Double> valueList = new ArrayList<>();
                for (int c = 0; c < sequence.getSizeC(); c++) {
                    valueList.clear();
                    int cntValues = 0;
                    double sumIntensity = 0;
                    double minIntensity = 0;
                    double maxIntensity = 0;
                    double sumSqIntensity = 0;
                    double sumBackgroundIntensity = 0;
                    int cntBackgroundValues = 0;

                    final int minX = (int) Math.max(0, d.getX() - outterDiskRadius);
                    final int minY = (int) Math.max(0, d.getY() - outterDiskRadius);
                    final int minZ = (int) Math.max(0, d.getZ() - diskRadius);

                    final int maxX = (int) Math.min(width - 1, Math.ceil(d.getX() + outterDiskRadius));
                    final int maxY = (int) Math.min(height - 1, Math.ceil(d.getY() + outterDiskRadius));
                    final int maxZ = (int) Math.min(depth - 1, Math.ceil(d.getZ() + diskRadius));

                    for (int z = minZ; z <= maxZ; z++) {
                        final double[] imageArray = imageTab[z][c];
                        for (int y = minY; y <= maxY; y++)
                            for (int x = minX; x <= maxX; x++) {
                                if ((d.getX() - x) * (d.getX() - x) + (d.getY() - y) * (d.getY() - y) + (d.getZ() - z) * (d.getZ() - z) <= diskRadius * diskRadius) {
                                    cntValues++;
                                    final double value = imageArray[x + y * width];
                                    sumIntensity += value;
                                    if (cntValues == 1) {
                                        minIntensity = value;
                                        maxIntensity = value;
                                    }
                                    else {
                                        if (value < minIntensity)
                                            minIntensity = value;
                                        if (value > maxIntensity)
                                            maxIntensity = value;
                                    }
                                    sumSqIntensity += value * value;
                                    valueList.add(value);
                                }
                                else if ((d.getX() - x) * (d.getX() - x) + (d.getY() - y) * (d.getY() - y) + (d.getZ() - z) * (d.getZ() - z) <= outterDiskRadius * outterDiskRadius) {
                                    cntBackgroundValues++;
                                    sumBackgroundIntensity += imageArray[x + y * width];
                                }
                            }
                        double meanIntensity = 0;
                        double varIntensity = 0;
                        if (cntValues > 0) {
                            meanIntensity = sumIntensity / (double) cntValues;
                            varIntensity = sumSqIntensity / (double) cntValues - meanIntensity * meanIntensity;
                            meanIntensity -= sumBackgroundIntensity / (double) cntBackgroundValues;
                        }

                        this.meanIntensity[c].add(t, meanIntensity);
                        this.maxIntensity[c].add(t, maxIntensity);
                        this.minIntensity[c].add(t, minIntensity);
                        this.varIntensity[c].add(t, varIntensity);
                        this.sumIntensity[c].add(t, sumIntensity);
                        Arrays.sort(valueList.toArray());
                        if (cntValues == 1)
                            this.medianIntensity[c].add(t, valueList.getFirst());
                        else if (cntValues > 0) {
                            if (cntValues % 2 == 0) {
                                this.medianIntensity[c].add(t, (valueList.get(cntValues / 2)) + (valueList.get((cntValues / 2)) - 1) / 2);
                            }
                            else
                                this.medianIntensity[c].add(t, valueList.get(cntValues / 2));
                        }
                        else
                            this.medianIntensity[c].add(t, 0);
                    }
                }
            }
        }
    }
}
