/*
 * Copyright (c) 2010-2024. Institut Pasteur.
 *
 * This file is part of Icy.
 * Icy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Icy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Icy. If not, see <https://www.gnu.org/licenses/>.
 */

package plugins.nchenouard.trackprocessorintensityprofile;

import org.bioimageanalysis.icy.common.collection.array.ArrayUtil;
import org.bioimageanalysis.icy.common.geom.rectangle.Rectangle5D;
import org.bioimageanalysis.icy.model.sequence.Sequence;
import org.bioimageanalysis.icy.model.sequence.SequenceUtil;
import org.bioimageanalysis.icy.system.logging.IcyLogger;
import org.jfree.data.xy.XYSeries;
import plugins.fab.trackmanager.TrackSegment;
import plugins.nchenouard.spot.Detection;

public class TrackAnalysisGaussianFit extends TrackAnalysis {
    XYSeries[] gaussianAmplitude;
    XYSeries[] gaussianBackground;

    TrackAnalysisGaussianFit(
            final TrackSegment track,
            final Sequence sequence,
            final String description,
            final AveragingType averagingMethod
    ) {
        super(track, sequence, description, averagingMethod);

        gaussianAmplitude = new XYSeries[sequence.getSizeC()];
        for (int c = 0; c < sequence.getSizeC(); c++)
            gaussianAmplitude[c] = new XYSeries(description);
        gaussianBackground = new XYSeries[sequence.getSizeC()];
        for (int c = 0; c < sequence.getSizeC(); c++)
            gaussianBackground[c] = new XYSeries(description);
    }

    @Override
    public void clearSeries() {
        for (final XYSeries xySeries : gaussianAmplitude) xySeries.clear();
        for (final XYSeries xySeries : gaussianBackground) xySeries.clear();
    }

    public void fillGaussianFitSeriesAtT(final int t, final double[][] imageTab, final double stdGaussian) {
        if (track.getFirstDetection().getT() <= t && track.getLastDetection().getT() >= t) {
            final Detection d = track.getDetectionAtTime(t);
            if (d != null) {
                final int z = (int) Math.round(d.getZ());
                final int width = sequence.getSizeX();
                final int height = sequence.getSizeY();
                // compute the Gaussian amplitude
                int minX = (int) d.getX() - halfCropSize;
                double pCropX = halfCropSize + (d.getX() - (int) d.getX());
                if (minX < 0) {
                    pCropX -= minX;
                    minX = 0;
                }
                int minY = (int) d.getY() - halfCropSize;
                double pCropY = halfCropSize + (d.getY() - (int) d.getY());
                if (minY < 0) {
                    pCropY -= minY;
                    minY = 0;
                }
                final int maxX = (int) Math.min(Math.ceil(d.getX()) + halfCropSize, width - 1);
                final int maxY = (int) Math.min(Math.ceil(d.getY()) + halfCropSize, height - 1);
                final int cropWidth = maxX - minX + 1;
                final int cropHeight = maxY - minY + 1;

                for (int c = 0; c < sequence.getSizeC(); c++) {
                    // fit the model background + amplitude*GaussianProfile to the cropped image around the detected position
                    //					Sequence crop = SequenceUtil.getSubSequence(sequence, minX, minY, z, t, cropWidth, cropHeight, 1, 1);
                    if (cropWidth <= 0 || cropHeight <= 0) {
                        IcyLogger.warn(this.getClass(), "Detection outside of image.");
                        this.gaussianAmplitude[c].add(t, 0);
                        this.gaussianBackground[c].add(t, 0);
                    }
                    else {
                        final Sequence crop = SequenceUtil.getSubSequence(
                                sequence,
                                new Rectangle5D.Integer(
                                        minX,
                                        minY,
                                        z,
                                        t,
                                        c,
                                        cropWidth,
                                        cropHeight,
                                        1,
                                        1,
                                        1
                                )
                        );
                        if (crop != null) {
                            final double[] cropImage = (double[]) ArrayUtil.arrayToDoubleArray(crop.getImage(0, 0, 0).getDataXY(0), crop.isSignedDataType());
                            // build the series of exponential sum sum exp(-((x - minX)^2 + (y - minY)^2)/(2*standardDeviation^2))
                            double sumExp = 0;
                            double sumExp2 = 0;
                            double sumExpVal = 0;
                            double sumVal = 0;
                            final double varGaussian2 = 2 * stdGaussian * stdGaussian;
                            for (int y = 0; y < cropHeight; y++)
                                for (int x = 0; x < cropWidth; x++) {
                                    final double valExp = Math.exp(-((y - pCropY) * (y - pCropY) + (x - pCropX) * (x - pCropX)) / varGaussian2);
                                    sumExp += valExp;
                                    sumExp2 += (valExp * valExp);
                                    final double val = cropImage[x + y * cropWidth];
                                    sumVal += val;
                                    sumExpVal += (val * valExp);
                                }
                            final double detMat = sumExp2 * cropHeight * cropWidth - sumExp * sumExp;
                            final double a11 = cropHeight * cropWidth / detMat;
                            final double a12 = -sumExp / detMat;
                            final double a21 = a12;
                            final double a22 = sumExp2 / detMat;

                            final double amplitude = a11 * sumExpVal + a12 * sumVal;
                            final double background = a21 * sumExpVal + a22 * sumVal;

                            this.gaussianAmplitude[c].add(t, amplitude);
                            this.gaussianBackground[c].add(t, background);
                        }
                        else {
                            this.gaussianAmplitude[c].add(t, 0);
                            this.gaussianBackground[c].add(t, 0);
                        }
                    }
                }
            }

        }
    }
}
