/*
 * Copyright (c) 2010-2024. Institut Pasteur.
 *
 * This file is part of Icy.
 * Icy is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Icy is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Icy. If not, see <https://www.gnu.org/licenses/>.
 */

package plugins.nchenouard.trackprocessorintensityprofile;

import org.apache.poi.ss.usermodel.Sheet;
import org.apache.poi.ss.usermodel.Workbook;
import org.bioimageanalysis.icy.common.collection.array.ArrayUtil;
import org.bioimageanalysis.icy.extension.plugin.annotation_.IcyPluginName;
import org.bioimageanalysis.icy.gui.dialog.SaveDialog;
import org.bioimageanalysis.icy.io.FileUtil;
import org.bioimageanalysis.icy.io.xls.XLSXUtil;
import org.bioimageanalysis.icy.model.sequence.Sequence;
import org.bioimageanalysis.icy.system.logging.IcyLogger;
import org.jetbrains.annotations.NotNull;
import org.jfree.chart.ChartFactory;
import org.jfree.chart.ChartPanel;
import org.jfree.chart.JFreeChart;
import org.jfree.chart.plot.PlotOrientation;
import org.jfree.data.xy.XYSeries;
import org.jfree.data.xy.XYSeriesCollection;
import plugins.fab.trackmanager.PluginTrackManagerProcessor;
import plugins.fab.trackmanager.TrackGroup;
import plugins.fab.trackmanager.TrackSegment;
import plugins.nchenouard.spot.Detection;

import javax.swing.*;
import java.awt.*;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Objects;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

/**
 * A TrackProcessor to monitor detection fluorescence along time
 *
 * @author nicolas chenouard
 * <p>
 * date 11/03/2012
 */
@IcyPluginName("Track Intensity Profile Processor")
public class TrackProcessorIntensityProfile extends PluginTrackManagerProcessor {
    //TODO: disable analysis thread when processor is disabled
    ArrayList<JFreeChart> charts;
    JPanel chartsPanel = new JPanel();
    JCheckBox displayLegendBox = new JCheckBox("Display legend");

    JPanel averagingDiskPanel;
    JPanel averagingGaussianFitPanel;
    JPanel averagingDiskNoBackgroundPanel;
    JPanel averagingMaskPanel;

    JComboBox<AveragingType> averagingTypeBox;
    CardLayout methodCardLayout;
    JPanel averagingMethodPanel;

    String[] criteria = new String[]{
            "Mean intensity",
            "Max intensity",
            "Min intensity",
            "Median intensity",
            "Intensity sum",
            "Intensity variance"
    };
    JComboBox<String> criteriaBox = new JComboBox<>(criteria);
    JComboBox<String> criteriaBox2 = new JComboBox<>(criteria);
    JComboBox<String> criteriaBox3 = new JComboBox<>(criteria);

    SpinnerNumberModel diskDiameterModel = new SpinnerNumberModel(2, 0, 1000, 0.5);
    SpinnerNumberModel diskOutterDiameterModel = new SpinnerNumberModel(4, 2, 1000, 0.5);

    String[] criteriaGaussian = new String[]{"Amplitude", "Background"};
    JComboBox<String> criteriaGaussianBox = new JComboBox<>(criteriaGaussian);
    JTextField standardDeviationTF;

    ArrayList<TrackAnalysis> analyzedTracks = new ArrayList<>();
    ArrayList<TrackAnalysis> toDisplayTracks = new ArrayList<>();

    ArrayList<int[]> positions = new ArrayList<>();
    ReentrantLock positionsLock = new ReentrantLock();

    ReentrantLock analyzedTracksLock = new ReentrantLock();
    ArrayList<TrackSegment> seriesToFill = new ArrayList<>();

    ReentrantLock seriesToFillLock = new ReentrantLock();
    Condition seriesToFillCondition = seriesToFillLock.newCondition();

    FillSeriesThread fillSeriesThread;

    Sequence selectedSequence = null;

    public TrackProcessorIntensityProfile() {
        this.setName("Intensity profile");

        this.panel.setLayout(new BorderLayout());

        final JPanel northPanel = new JPanel();
        northPanel.setLayout(new BorderLayout());

        final JPanel averagingTypePanel = new JPanel(new GridLayout(2, 1));
        averagingTypePanel.add(new JLabel("Averaging type"));
        averagingTypeBox = new JComboBox<>(AveragingType.values());
        averagingTypeBox.addActionListener(e -> {
            methodCardLayout.show(averagingMethodPanel, Objects.requireNonNull(averagingTypeBox.getSelectedItem()).toString());
            refreshAllResults();
            updateDisplay();
        });
        averagingTypePanel.add(averagingTypeBox);
        northPanel.add(averagingTypePanel, BorderLayout.NORTH);

        methodCardLayout = new CardLayout();
        averagingMethodPanel = new JPanel(methodCardLayout);
        northPanel.add(averagingMethodPanel, BorderLayout.CENTER);

        averagingDiskPanel = new JPanel(new GridLayout(2, 2));
        averagingDiskPanel.add(new JLabel("Diameter of the disk area in pixels"));
        final JSpinner diskDiameterSpinner = new JSpinner(diskDiameterModel);
        diskDiameterSpinner.addChangeListener(e -> {
            //updatePositions();
            refreshAllResults();
        });
        averagingDiskPanel.add(diskDiameterSpinner);
        averagingDiskPanel.add(new JLabel("Criterion to display"));
        averagingDiskPanel.add(criteriaBox);
        criteriaBox.addActionListener(e -> updateDisplay());
        averagingMethodPanel.add(averagingDiskPanel, AveragingType.DISK.toString());

        averagingDiskNoBackgroundPanel = new JPanel(new GridLayout(4, 2));
        averagingDiskNoBackgroundPanel.add(new JLabel("Diameter of the disk area in pixels"));
        final JSpinner diskDiameterSpinner2 = new JSpinner(diskDiameterModel);
        diskDiameterSpinner2.addChangeListener(e -> refreshAllResults());
        averagingDiskNoBackgroundPanel.add(diskDiameterSpinner2);
        final JSpinner diskOutterDiameterSpinner2 = new JSpinner(diskOutterDiameterModel);
        diskOutterDiameterSpinner2.addChangeListener(e -> refreshAllResults());
        averagingDiskNoBackgroundPanel.add(new JLabel("Diameter of the background disk area in pixels"));
        averagingDiskNoBackgroundPanel.add(diskOutterDiameterSpinner2);
        averagingDiskNoBackgroundPanel.add(new JLabel("Criterion to display"));
        averagingDiskNoBackgroundPanel.add(criteriaBox2);
        criteriaBox2.addActionListener(e -> updateDisplay());
        averagingMethodPanel.add(averagingDiskNoBackgroundPanel, AveragingType.DISK_BACKGROUND_CORRECTED.toString());

        averagingMaskPanel = new JPanel(new GridLayout(2, 2));
        averagingMaskPanel.add(new JLabel("Missing detections masks replaced by value at center of mass."));
        averagingMaskPanel.add(new JLabel(""));
        averagingMaskPanel.add(new JLabel("Criterion to display"));
        averagingMaskPanel.add(criteriaBox3);
        criteriaBox3.addActionListener(e -> updateDisplay());
        averagingMethodPanel.add(averagingMaskPanel, AveragingType.SPOT_MASK.toString());

        averagingGaussianFitPanel = new JPanel(new GridLayout(2, 2));
        averagingGaussianFitPanel.add(new JLabel("Standard deviation"));
        standardDeviationTF = new JTextField("1.5", 5);
        standardDeviationTF.addActionListener(e -> updateDisplay());
        averagingGaussianFitPanel.add(standardDeviationTF);
        averagingGaussianFitPanel.add(new JLabel("Criterion to display"));
        averagingGaussianFitPanel.add(criteriaGaussianBox);
        criteriaGaussianBox.addActionListener(e -> updateDisplay());
        averagingMethodPanel.add(averagingGaussianFitPanel, AveragingType.GAUSSIAN_FIT2D.toString());

        northPanel.add(displayLegendBox, BorderLayout.SOUTH);
        displayLegendBox.addActionListener(e -> updateDisplay());

        this.panel.add(northPanel, BorderLayout.NORTH);

        final JScrollPane centerPane = new JScrollPane(chartsPanel);
        this.panel.add(centerPane, BorderLayout.CENTER);

        final JPanel southPane = new JPanel();
        southPane.setLayout(new BorderLayout());
        final JButton exportResultsButton = new JButton("Export results");
        exportResultsButton.addActionListener(e -> exportResults());
        southPane.add(exportResultsButton);

        this.panel.add(southPane, BorderLayout.SOUTH);

        fillSeriesThread = new FillSeriesThread();
        fillSeriesThread.start();

        changeSelectedSequence(getActiveSequence());
    }

    public void exportResults() {
        final ArrayList<TrackAnalysis> analyzedTracks = new ArrayList<>();
        try {
            seriesToFillLock.lock();
            analyzedTracks.addAll(this.analyzedTracks);
        }
        finally {
            seriesToFillLock.unlock();
        }
        final Object[] possibilities = {"Excel (xlsx)", "Text (txt)"};
        final String s = (String) JOptionPane.showInputDialog(
                this.getPanel(), "Select the output format",
                "Intensity output",
                JOptionPane.PLAIN_MESSAGE,
                null,
                possibilities,
                "Excel (xlsx)");
        //If a string was returned, say so.
        if ((s != null) && (!s.isBlank())) {
            if (s.equals(("Excel (xlsx)")))
                exportToXLS(analyzedTracks, (AveragingType) averagingTypeBox.getSelectedItem());
            else
                exportToTXT(analyzedTracks, (AveragingType) averagingTypeBox.getSelectedItem());
        }
    }

    private void exportToTXT(final ArrayList<TrackAnalysis> trackList, final AveragingType averType) {
        final JFileChooser chooser = new JFileChooser();
        chooser.setDialogTitle("Choose txt file.");
        final int returnVal = chooser.showOpenDialog(this.panel);
        if (returnVal != JFileChooser.APPROVE_OPTION) return;

        File file = chooser.getSelectedFile();
        if (file == null)
            return;
        if (!file.getName().endsWith(".txt")) {
            try {
                file = new File(file.getCanonicalPath() + ".txt");
            }
            catch (final IOException e) {
                IcyLogger.error(this.getClass(), e);
                return;
            }
        }
        final ArrayList<TrackAnalysis> tList = new ArrayList<>();
        for (final TrackAnalysis ta : trackList)
            if (ta.averagingType == averType)
                tList.add(ta);
        final int numChannels = computeMaxNumChannels(tList);
        final File[] outputFiles = new File[numChannels];
        if (numChannels > 1)
            try {
                for (int c = 0; c < numChannels; c++)
                    outputFiles[c] = new File(file.getCanonicalPath().substring(0, file.getCanonicalPath().toCharArray().length - 4) + "_" + c + ".txt");
            }
            catch (final IOException e) {
                IcyLogger.error(this.getClass(), e, "Could not create output file.");
                return;
            }
        else
            outputFiles[0] = file;
        final File startIdxFile; // = null;
        try {
            startIdxFile = new File(file.getCanonicalPath().substring(0, file.getCanonicalPath().toCharArray().length - 4) + "_trackStartIndex.txt");
        }
        catch (final IOException e1) {
            IcyLogger.error(this.getClass(), e1, "Could not create start index file.");
            return;
        }
        final File startFile = startIdxFile;

        if (tList.isEmpty())
            return;
        final Thread savingThread = new Thread() {
            @Override
            public void run() {
                final BufferedWriter[] writers = new BufferedWriter[numChannels];
                try {
                    for (int i = 0; i < numChannels; i++)
                        writers[i] = new BufferedWriter(new FileWriter(outputFiles[i]));
                }
                catch (final IOException e) {
                    IcyLogger.error(this.getClass(), e, "Could not create output file.");
                    return;
                }

                IcyLogger.info(this.getClass(), "Saving results.");
                final ArrayList<Integer> startIdx = new ArrayList<>();
                switch (averType) {
                    case DISK:
                        try {
                            // append tracks one after another in the same file
                            // then create a file for indices of start and end of tracks
                            int k = 0;
                            for (final TrackAnalysis ta : tList) {
                                startIdx.add(k);
                                final TrackAnalysisDiskAveraging taa = (TrackAnalysisDiskAveraging) ta;
                                for (int c = 0; c < numChannels; c++) {
                                    for (int i = 0; i < taa.meanIntensity[c].getItemCount(); i++) {
                                        final int t = taa.meanIntensity[c].getX(i).intValue();
                                        final double mean = taa.meanIntensity[c].getY(i).doubleValue();
                                        final double max = taa.maxIntensity[c].getY(i).doubleValue();
                                        final double min = taa.minIntensity[c].getY(i).doubleValue();
                                        final double var = taa.varIntensity[c].getY(i).doubleValue();
                                        final double median = taa.medianIntensity[c].getY(i).doubleValue();
                                        writers[c].write(t + ", " + mean + ", " + min + ", " + max + ", " + var + ", " + median + "\n");
                                        if (c == 0)
                                            k++;
                                    }
                                }
                            }
                        }
                        catch (final IOException e) {
                            IcyLogger.error(this.getClass(), e);
                            return;
                        }
                        break;
                    case SPOT_MASK:
                        try {
                            // append tracks one after another in the same file
                            // then create a file for indices of start and end of tracks
                            int k = 0;
                            for (final TrackAnalysis ta : tList) {
                                startIdx.add(k);
                                final TrackAnalysisMaskAveraging taa = (TrackAnalysisMaskAveraging) ta;
                                for (int c = 0; c < numChannels; c++) {
                                    for (int i = 0; i < taa.meanIntensity[c].getItemCount(); i++) {
                                        final int t = taa.meanIntensity[c].getX(i).intValue();
                                        final double mean = taa.meanIntensity[c].getY(i).doubleValue();
                                        final double max = taa.maxIntensity[c].getY(i).doubleValue();
                                        final double min = taa.minIntensity[c].getY(i).doubleValue();
                                        final double var = taa.varIntensity[c].getY(i).doubleValue();
                                        final double median = taa.medianIntensity[c].getY(i).doubleValue();
                                        writers[c].write(t + ", " + mean + ", " + min + ", " + max + ", " + var + ", " + median + "\n");
                                        if (c == 0)
                                            k++;
                                    }
                                }
                            }
                        }
                        catch (final IOException e) {
                            IcyLogger.error(this.getClass(), e);
                            return;
                        }
                        break;
                    case GAUSSIAN_FIT2D:
                        try {
                            // append tracks one after another in the same file
                            // then create a file for indices of start and end of tracks
                            int k = 0;
                            for (final TrackAnalysis ta : tList) {
                                startIdx.add(k);
                                final TrackAnalysisGaussianFit taa = (TrackAnalysisGaussianFit) ta;
                                for (int c = 0; c < numChannels; c++) {
                                    for (int i = 0; i < taa.gaussianAmplitude[c].getItemCount(); i++) {
                                        final int t = taa.gaussianAmplitude[c].getX(i).intValue();
                                        final double amplitude = taa.gaussianAmplitude[c].getY(i).doubleValue();
                                        final double background = taa.gaussianAmplitude[c].getY(i).doubleValue();
                                        writers[c].write(t + ", " + amplitude + ", " + background + "\n");
                                        if (c == 0)
                                            k++;
                                    }
                                }
                            }
                        }
                        catch (final IOException e) {
                            IcyLogger.error(this.getClass(), e);
                            return;
                        }
                        break;
                    case DISK_BACKGROUND_CORRECTED:
                        try {
                            // append tracks one after another in the same file
                            // then create a file for indices of start and end of tracks
                            int k = 0;
                            for (final TrackAnalysis ta : tList) {
                                startIdx.add(k);
                                final TrackAnalysisDiskAveragingNoBackground taa = (TrackAnalysisDiskAveragingNoBackground) ta;
                                for (int c = 0; c < numChannels; c++) {
                                    for (int i = 0; i < taa.meanIntensity[c].getItemCount(); i++) {
                                        final int t = taa.meanIntensity[c].getX(i).intValue();
                                        final double mean = taa.meanIntensity[c].getY(i).doubleValue();
                                        final double max = taa.maxIntensity[c].getY(i).doubleValue();
                                        final double min = taa.minIntensity[c].getY(i).doubleValue();
                                        final double var = taa.varIntensity[c].getY(i).doubleValue();
                                        final double median = taa.medianIntensity[c].getY(i).doubleValue();
                                        writers[c].write(t + ", " + mean + ", " + min + ", " + max + ", " + var + ", " + median + "\n");
                                        if (c == 0)
                                            k++;
                                    }
                                }
                            }
                        }
                        catch (final IOException e) {
                            IcyLogger.error(this.getClass(), e);
                            return;
                        }
                        break;
                    default:
                        break;
                }
                try {
                    for (final BufferedWriter bwr : writers)
                        bwr.close();
                }
                catch (final IOException e) {
                    IcyLogger.error(this.getClass(), e, "Could not close output file.");
                    return;
                }
                final BufferedWriter bwr;
                try {
                    bwr = new BufferedWriter(new FileWriter(startFile));
                    for (final Integer it : startIdx)
                        bwr.write(it.toString() + "\n");
                    bwr.close();
                }
                catch (final IOException e) {
                    IcyLogger.error(this.getClass(), e);
                    return;
                }
                savingThreadList.remove(this);
                IcyLogger.success(this.getClass(), "Results saved.");
            }
        };
        savingThreadList.add(savingThread);
        savingThread.start();
    }

    private void exportToXLS(final ArrayList<TrackAnalysis> trackAnalysisList, final AveragingType averType) {
        final String filename = SaveDialog.chooseFile(
                "Choose XLSX file",
                null,
                "track-intensity-profile",
                XLSXUtil.FILE_DOT_EXTENSION
        );

        if (filename == null)
            return;

        if (!FileUtil.getFileExtension(filename, false).equalsIgnoreCase(XLSXUtil.FILE_EXTENSION))
            return;

        final File XLSFile = new File(filename);

        final ArrayList<TrackAnalysis> trackList = new ArrayList<>();
        for (final TrackAnalysis ta : trackAnalysisList)
            if (ta.averagingType == averType)
                trackList.add(ta);
        if (trackList.isEmpty())
            return;

        final Thread savingThread = new Thread() {
            @Override
            public void run() {
                IcyLogger.info(this.getClass(), "Saving results");
                // save disk averaging results first

                {
                    final Workbook workbook;
                    try {
                        if (XLSFile.exists())
                            workbook = XLSXUtil.loadWorkbookForWrite(XLSFile);
                        else
                            workbook = XLSXUtil.createWorkbook();
                    }
                    catch (final IOException e) {
                        IcyLogger.error(this.getClass(), e, "Cannot open file.");
                        return;
                    }

                    final int numChannels = computeMaxNumChannels(trackList);
                    switch (averType) {
                        case DISK: {
                            //int s = 0;
                            for (int c = 0; c < numChannels; c++) {
                                for (int i = 0; i < 6; i++) {
                                    final Sheet sheet = XLSXUtil.getPage(workbook, "Channel " + c + " - " + criteria[i]);

                                    // create the time axis
                                    int maxT = 0;
                                    int minT = Integer.MAX_VALUE;
                                    for (final TrackAnalysis trkAnalysis : trackList) {
                                        if (!trkAnalysis.track.getDetectionList().isEmpty()) {
                                            if (trkAnalysis.track.getLastDetection().getT() > maxT)
                                                maxT = trkAnalysis.track.getLastDetection().getT();
                                            if (trkAnalysis.track.getFirstDetection().getT() < minT)
                                                minT = trkAnalysis.track.getFirstDetection().getT();
                                        }
                                    }
                                    XLSXUtil.setCellString(sheet, 0, 0, "Frame number");
                                    for (int t = minT; t <= maxT; t++)
                                        XLSXUtil.setCellNumber(sheet, 0, t - minT + 1, t);

                                    int cntTrack = 0;
                                    for (final TrackAnalysis trka : trackList) {
                                        final TrackAnalysisDiskAveraging trkAnalysis = (TrackAnalysisDiskAveraging) trka;
                                        if (!trkAnalysis.track.getDetectionList().isEmpty()) {
                                            cntTrack++;

                                            XLSXUtil.setCellString(sheet, cntTrack, 0, trkAnalysis.description);
                                            final int firstT = trkAnalysis.track.getFirstDetection().getT();
                                            switch (i) {
                                                case 0: {
                                                    if (trkAnalysis.meanIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.meanIntensity[c].getItemCount(); j++) {
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.meanIntensity[c].getY(j).doubleValue());
                                                        }
                                                    break;
                                                }
                                                case 1:
                                                    if (trkAnalysis.maxIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.maxIntensity[c].getItemCount(); j++) {
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.maxIntensity[c].getY(j).doubleValue());
                                                        }
                                                    break;
                                                case 2:
                                                    if (trkAnalysis.minIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.minIntensity[c].getItemCount(); j++) {
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.minIntensity[c].getY(j).doubleValue());
                                                        }
                                                    break;
                                                case 3:
                                                    if (trkAnalysis.medianIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.medianIntensity[c].getItemCount(); j++) {
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.medianIntensity[c].getY(j).doubleValue());
                                                        }
                                                    break;
                                                case 4:
                                                    if (trkAnalysis.sumIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.sumIntensity[c].getItemCount(); j++) {
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.sumIntensity[c].getY(j).doubleValue());
                                                        }
                                                    break;
                                                case 5:
                                                    if (trkAnalysis.varIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.varIntensity[c].getItemCount(); j++) {
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.varIntensity[c].getY(j).doubleValue());
                                                        }
                                                    break;
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }
                        case SPOT_MASK: {
                            for (int c = 0; c < numChannels; c++) {
                                for (int i = 0; i < 6; i++) {
                                    final Sheet sheet = XLSXUtil.getPage(workbook, "Channel " + c + " - " + criteria[i]);

                                    // create the time axis
                                    int maxT = 0;
                                    int minT = Integer.MAX_VALUE;
                                    for (final TrackAnalysis trkAnalysis : trackList) {
                                        if (!trkAnalysis.track.getDetectionList().isEmpty()) {
                                            if (trkAnalysis.track.getLastDetection().getT() > maxT)
                                                maxT = trkAnalysis.track.getLastDetection().getT();
                                            if (trkAnalysis.track.getFirstDetection().getT() < minT)
                                                minT = trkAnalysis.track.getFirstDetection().getT();
                                        }
                                    }
                                    XLSXUtil.setCellString(sheet, 0, 0, "Frame number");

                                    for (int t = minT; t <= maxT; t++)
                                        XLSXUtil.setCellNumber(sheet, 0, t - minT + 1, t);

                                    int cntTrack = 0;
                                    for (final TrackAnalysis trka : trackList) {
                                        final TrackAnalysisMaskAveraging trkAnalysis = (TrackAnalysisMaskAveraging) trka;
                                        if (!trkAnalysis.track.getDetectionList().isEmpty()) {
                                            cntTrack++;

                                            XLSXUtil.setCellString(sheet, cntTrack, 0, trkAnalysis.description);
                                            final int firstT = trkAnalysis.track.getFirstDetection().getT();
                                            switch (i) {
                                                case 0: {
                                                    if (trkAnalysis.meanIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.meanIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.meanIntensity[c].getY(j).doubleValue());
                                                    break;
                                                }
                                                case 1:
                                                    if (trkAnalysis.maxIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.maxIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.maxIntensity[c].getY(j).doubleValue());
                                                    break;
                                                case 2:
                                                    if (trkAnalysis.minIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.minIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.minIntensity[c].getY(j).doubleValue());
                                                    break;
                                                case 3:
                                                    if (trkAnalysis.medianIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.medianIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.medianIntensity[c].getY(j).doubleValue());
                                                    break;
                                                case 4:
                                                    if (trkAnalysis.sumIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.sumIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.sumIntensity[c].getY(j).doubleValue());
                                                    break;
                                                case 5:
                                                    if (trkAnalysis.varIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.varIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.varIntensity[c].getY(j).doubleValue());
                                                    break;
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }
                        case GAUSSIAN_FIT2D: {
                            for (int c = 0; c < numChannels; c++) {
                                for (int i = 0; i < 2; i++) {
                                    final Sheet sheet = XLSXUtil.getPage(workbook, "Channel " + c + " - " + criteriaGaussian[i]);

                                    // create the time axis
                                    int maxT = 0;
                                    int minT = Integer.MAX_VALUE;
                                    for (final TrackAnalysis trkAnalysis : trackList) {
                                        if (!trkAnalysis.track.getDetectionList().isEmpty()) {
                                            if (trkAnalysis.track.getLastDetection().getT() > maxT)
                                                maxT = trkAnalysis.track.getLastDetection().getT();
                                            if (trkAnalysis.track.getFirstDetection().getT() < minT)
                                                minT = trkAnalysis.track.getFirstDetection().getT();
                                        }
                                    }
                                    XLSXUtil.setCellString(sheet, 0, 0, "Frame number");

                                    for (int t = minT; t <= maxT; t++)
                                        XLSXUtil.setCellNumber(sheet, 0, t - minT + 1, t);

                                    int cntTrack = 0;
                                    for (final TrackAnalysis trka : trackList) {
                                        final TrackAnalysisGaussianFit trkAnalysis = (TrackAnalysisGaussianFit) trka;
                                        if (!trkAnalysis.track.getDetectionList().isEmpty()) {
                                            cntTrack++;

                                            XLSXUtil.setCellString(sheet, cntTrack, 0, trkAnalysis.description);
                                            final int firstT = trkAnalysis.track.getFirstDetection().getT();
                                            switch (i) {
                                                case 0: {
                                                    if (trkAnalysis.gaussianAmplitude.length >= c)
                                                        for (int j = 0; j < trkAnalysis.gaussianAmplitude[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.gaussianAmplitude[c].getY(j).doubleValue());
                                                    break;
                                                }
                                                case 1:
                                                    if (trkAnalysis.gaussianBackground.length >= c)
                                                        for (int j = 0; j < trkAnalysis.gaussianBackground[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.gaussianBackground[c].getY(j).doubleValue());
                                                    break;
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }
                        case DISK_BACKGROUND_CORRECTED: {
                            for (int c = 0; c < numChannels; c++) {
                                for (int i = 0; i < 6; i++) {
                                    final Sheet sheet = XLSXUtil.getPage(workbook, "Channel " + c + " - " + criteria[i]);

                                    // create the time axis
                                    int maxT = 0;
                                    int minT = Integer.MAX_VALUE;
                                    for (final TrackAnalysis trkAnalysis : trackList) {
                                        if (!trkAnalysis.track.getDetectionList().isEmpty()) {
                                            if (trkAnalysis.track.getLastDetection().getT() > maxT)
                                                maxT = trkAnalysis.track.getLastDetection().getT();
                                            if (trkAnalysis.track.getFirstDetection().getT() < minT)
                                                minT = trkAnalysis.track.getFirstDetection().getT();
                                        }
                                    }
                                    XLSXUtil.setCellString(sheet, 0, 0, "Frame number");

                                    for (int t = minT; t <= maxT; t++)
                                        XLSXUtil.setCellNumber(sheet, 0, t - minT + 1, t);

                                    int cntTrack = 0;
                                    for (final TrackAnalysis trka : trackList) {
                                        final TrackAnalysisDiskAveragingNoBackground trkAnalysis = (TrackAnalysisDiskAveragingNoBackground) trka;
                                        if (!trkAnalysis.track.getDetectionList().isEmpty()) {
                                            cntTrack++;

                                            XLSXUtil.setCellString(sheet, cntTrack, 0, trkAnalysis.description);
                                            final int firstT = trkAnalysis.track.getFirstDetection().getT();
                                            switch (i) {
                                                case 0: {
                                                    if (trkAnalysis.meanIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.meanIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.meanIntensity[c].getY(j).doubleValue());
                                                    break;
                                                }
                                                case 1:
                                                    if (trkAnalysis.maxIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.maxIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.maxIntensity[c].getY(j).doubleValue());
                                                    break;
                                                case 2:
                                                    if (trkAnalysis.minIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.minIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.minIntensity[c].getY(j).doubleValue());
                                                    break;
                                                case 3:
                                                    if (trkAnalysis.medianIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.medianIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.medianIntensity[c].getY(j).doubleValue());
                                                    break;
                                                case 4:
                                                    if (trkAnalysis.sumIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.sumIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.sumIntensity[c].getY(j).doubleValue());
                                                    break;
                                                case 5:
                                                    if (trkAnalysis.varIntensity.length >= c)
                                                        for (int j = 0; j < trkAnalysis.varIntensity[c].getItemCount(); j++)
                                                            XLSXUtil.setCellNumber(sheet, cntTrack, firstT + j + 1, trkAnalysis.varIntensity[c].getY(j).doubleValue());
                                                    break;
                                            }
                                        }
                                    }
                                }
                            }
                            break;
                        }
                        default:
                            IcyLogger.error(this.getClass(), "Unknown averaging type. Export cancelled.");
                            return;
                    }
                    try {
                        IcyLogger.debug(this.getClass(), "trying to save in: " + XLSFile.getAbsolutePath());
                        XLSXUtil.saveAndClose(workbook, XLSFile);
                    }
                    catch (final IOException e) {
                        IcyLogger.error(this.getClass(), e, "Could not save track profile.");
                    }
                }
                savingThreadList.remove(this);
                IcyLogger.success(this.getClass(), "Results saved.");
            }
        };
        savingThreadList.add(savingThread);
        savingThread.start();
    }

    private final ArrayList<Thread> savingThreadList = new ArrayList<>();

    @Override
    public void Close() {
        savingThreadList.clear();
        analyzeThreadList.clear();
    }

    ArrayList<Thread> analyzeThreadList = new ArrayList<>();

    @Override
    public void Compute() {
        if (!isEnabled()) return;
        final Sequence sequence = this.trackPool.getDisplaySequence();
        if (sequence != selectedSequence) {
            changeSelectedSequence(sequence);
            return;
        }
        final AveragingType averagingMethod = (AveragingType) averagingTypeBox.getSelectedItem();
        final Thread analyzeThread = new Thread() {
            @Override
            public void run() {
                // then performAnalysis in separate Thread
                try {
                    seriesToFillLock.lock();
                    //check whether some tracks were removed
                    boolean notFound = false;
                    for (final TrackAnalysis ta : analyzedTracks)
                        if (!trackPool.getTrackSegmentList().contains(ta.track)) {
                            notFound = true;
                            break;
                        }
                    // if missing tracks, recompute everything as labels have changed for tracks
                    if (notFound)
                        analyzedTracks.clear();

                    // get the tracks to Analyze
                    final ArrayList<TrackSegment> toAnalyzeTracks = getTracksToAnalyze(averagingMethod);

                    for (final TrackSegment ts : toAnalyzeTracks) {
                        if (!seriesToFill.contains(ts))
                            seriesToFill.add(ts);
                    }
                    seriesToFillCondition.signalAll();
                }
                finally {
                    seriesToFillLock.unlock();
                    analyzeThreadList.remove(this);
                }
            }
        };
        analyzeThreadList.add(analyzeThread);
        analyzeThread.start();
    }

    private void changeSelectedSequence(final Sequence selectedSequence) {
        if (selectedSequence != this.selectedSequence) {
            //			if (this.selectedSequence!=null)
            //				this.selectedSequence.removePainter(this);
            this.selectedSequence = selectedSequence;
            //			this.selectedSequence.addPainter(this);
            refreshAllResults();
        }
    }

    public ArrayList<TrackSegment> getTracksToAnalyze(final AveragingType averagingMethod) {
        //check whether the enabled tracks have already been analyzed
        final ArrayList<TrackSegment> enabledTracks = new ArrayList<>();
        final ArrayList<TrackSegment> copyTracks = new ArrayList<>();
        if (trackPool != null)
            copyTracks.addAll(trackPool.getTrackSegmentList());
        for (final TrackSegment track : copyTracks) {
            boolean enabled = false;
            for (final Detection d : track.getDetectionList())
                if (d.isEnabled()) {
                    enabled = true;
                    break;
                }
            if (enabled)
                enabledTracks.add(track);
        }
        final ArrayList<TrackSegment> toAnalyzeTracks = new ArrayList<>();
        for (final TrackSegment ts : enabledTracks) {
            boolean found = false;
            for (final TrackAnalysis analysis : analyzedTracks) {
                if (analysis.track == ts && analysis.averagingType == averagingMethod) {
                    found = true;
                    break;
                }
            }
            if (!found)
                toAnalyzeTracks.add(ts);
        }
        return toAnalyzeTracks;
    }

    @Override
    public void displaySequenceChanged() {
        changeSelectedSequence(getActiveSequence());
    }

    public void refreshAllResults() {
        try {
            final AveragingType averagingMethod = (AveragingType) averagingTypeBox.getSelectedItem();
            seriesToFillLock.lock();
            analyzedTracks.clear();
            seriesToFill.clear();
            seriesToFill.addAll(getTracksToAnalyze(averagingMethod));
            seriesToFillCondition.signalAll();
        }
        finally {
            seriesToFillLock.unlock();
        }
    }

    class FillSeriesThread extends Thread {
        boolean stop = false;

        public FillSeriesThread() {
        }

        @Override
        public void run() {
            while (!stop) {
                try {
                    seriesToFillLock.lock();
                    if (seriesToFill.isEmpty())
                        seriesToFillCondition.await();
                    if (selectedSequence != null) {
                        final double stdGaussian = Double.parseDouble(standardDeviationTF.getText());
                        final double diskRadius = ((Number) (diskDiameterModel.getValue())).doubleValue();
                        final double diskOutterRadius = ((Number) (diskOutterDiameterModel.getValue())).doubleValue();


                        final AveragingType averagingMethod = (AveragingType) Objects.requireNonNull(averagingTypeBox.getSelectedItem());

                        fillSeries(seriesToFill, selectedSequence, stdGaussian, diskRadius, diskOutterRadius, averagingMethod);//TODO: should be outside of the lock
                    }
                    //TODO: if outside the lock, some new task may come, while already being processed
                    seriesToFill.clear();
                }
                catch (final InterruptedException e) {
                    IcyLogger.error(this.getClass(), e, "FillSeriesThread interrupted.");
                }
                finally {
                    seriesToFillLock.unlock();
                }
                SwingUtilities.invokeLater(TrackProcessorIntensityProfile.this::updateDisplay);
            }
        }
    }

    public void updateDisplay() {
        final ArrayList<TrackAnalysis> analyzedTrkList = new ArrayList<>();
        try {
            analyzedTracksLock.lock();
            analyzedTrkList.addAll(analyzedTracks);
        }
        finally {
            analyzedTracksLock.unlock();
        }
        updateCharts(analyzedTrkList, (AveragingType) averagingTypeBox.getSelectedItem());
    }

    private int computeMaxNumChannels(final @NotNull ArrayList<TrackAnalysis> analyzedTrkList) {
        int maxNumChannels = 0;
        for (final TrackAnalysis tr : analyzedTrkList)
            if (tr.sequence.getSizeC() > maxNumChannels)
                maxNumChannels = tr.sequence.getSizeC();
        return maxNumChannels;
    }

    private void updateCharts(final ArrayList<TrackAnalysis> analyzedTrkList, final AveragingType averagingMethod) {
        //	displayedTracks.clear();
        final int maxNumChannels = computeMaxNumChannels(analyzedTrkList);
        chartsPanel.removeAll();
        if (maxNumChannels > 0) {
            chartsPanel.setLayout(new GridLayout(maxNumChannels, 1));
        }
        else
            return;
        for (int c = 0; c < maxNumChannels; c++) {
            final XYSeriesCollection xyDataset = new XYSeriesCollection();
            for (final TrackAnalysis tr : analyzedTrkList) {
                if (averagingMethod == tr.averagingType) {
                    int firstT = -1;
                    int lastT = -1;
                    boolean nonEmpty = false;
                    if (tr.track.getDetectionList().size() == 1) {
                        if (tr.track.getDetectionList().getFirst().isEnabled()) {
                            firstT = tr.track.getDetectionList().getFirst().getT();
                            lastT = firstT;
                            nonEmpty = true;
                        }
                    }
                    else {
                        for (int t = tr.track.getFirstDetection().getT(); t <= tr.track.getLastDetection().getT(); t++)
                            if (tr.track.getDetectionAtTime(t).isEnabled()) {
                                firstT = t;
                                lastT = t;
                                nonEmpty = true;
                                break;
                            }
                        if (nonEmpty)
                            for (int t = firstT + 1; t <= tr.track.getLastDetection().getT(); t++)
                                if (tr.track.getDetectionAtTime(t).isEnabled())
                                    lastT = t;
                    }
                    if (nonEmpty) {
                        //if (!displayedTracks.contains(tr.track))
                        //	displayedTracks.add(tr);
                        firstT = firstT - tr.track.getFirstDetection().getT();
                        lastT = lastT - tr.track.getFirstDetection().getT();
                        try {
                            switch (tr.averagingType) {
                                case DISK: {
                                    final XYSeries cropSeries; // = null;
                                    switch (criteriaBox.getSelectedIndex()) {
                                        case 0:
                                            cropSeries = ((TrackAnalysisDiskAveraging) tr).meanIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 1:
                                            cropSeries = ((TrackAnalysisDiskAveraging) tr).maxIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 2:
                                            cropSeries = ((TrackAnalysisDiskAveraging) tr).minIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 3:
                                            cropSeries = ((TrackAnalysisDiskAveraging) tr).medianIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 4:
                                            cropSeries = ((TrackAnalysisDiskAveraging) tr).sumIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 5:
                                            cropSeries = ((TrackAnalysisDiskAveraging) tr).varIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        default:
                                            IcyLogger.error(this.getClass(), "Series not found.");
                                            return;
                                    }
                                    xyDataset.addSeries(cropSeries);
                                    break;
                                }
                                case SPOT_MASK: {
                                    XYSeries cropSeries = null;
                                    switch (criteriaBox3.getSelectedIndex()) {
                                        case 0:
                                            cropSeries = ((TrackAnalysisMaskAveraging) tr).meanIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 1:
                                            cropSeries = ((TrackAnalysisMaskAveraging) tr).maxIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 2:
                                            cropSeries = ((TrackAnalysisMaskAveraging) tr).minIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 3:
                                            cropSeries = ((TrackAnalysisMaskAveraging) tr).medianIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 4:
                                            cropSeries = ((TrackAnalysisMaskAveraging) tr).sumIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 5:
                                            cropSeries = ((TrackAnalysisMaskAveraging) tr).varIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        default:
                                            IcyLogger.error(this.getClass(), "Series not found.");
                                    }
                                    xyDataset.addSeries(cropSeries);
                                    break;
                                }
                                case GAUSSIAN_FIT2D: {
                                    XYSeries cropSeries = null;
                                    switch (criteriaGaussianBox.getSelectedIndex()) {
                                        case 0:
                                            cropSeries = ((TrackAnalysisGaussianFit) tr).gaussianAmplitude[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 1:
                                            cropSeries = ((TrackAnalysisGaussianFit) tr).gaussianBackground[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                    }
                                    xyDataset.addSeries(cropSeries);
                                    break;
                                }
                                case DISK_BACKGROUND_CORRECTED: {

                                    XYSeries cropSeries = null;
                                    switch (criteriaBox2.getSelectedIndex()) {
                                        case 0:
                                            cropSeries = ((TrackAnalysisDiskAveragingNoBackground) tr).meanIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 1:
                                            cropSeries = ((TrackAnalysisDiskAveragingNoBackground) tr).maxIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 2:
                                            cropSeries = ((TrackAnalysisDiskAveragingNoBackground) tr).minIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 3:
                                            cropSeries = ((TrackAnalysisDiskAveragingNoBackground) tr).medianIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 4:
                                            cropSeries = ((TrackAnalysisDiskAveragingNoBackground) tr).sumIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        case 5:
                                            cropSeries = ((TrackAnalysisDiskAveragingNoBackground) tr).varIntensity[c].createCopy(firstT, lastT);
                                            cropSeries.setDescription(tr.description);
                                            break;
                                        default:
                                            IcyLogger.error(this.getClass(), "Series not found.");
                                    }
                                    xyDataset.addSeries(cropSeries);
                                    break;
                                }
                                default:
                                    break;
                            }
                        }
                        catch (final CloneNotSupportedException e) {
                            IcyLogger.error(this.getClass(), e);
                        }
                    }
                }
            }

            final String TitleString = "Channel " + c;
            final String TitleString2 = "Frame";
            final String TitleString3 = "Intensity criterion";
            final boolean displayLegend = displayLegendBox.isSelected();

            final JFreeChart chart = ChartFactory.createXYLineChart(
                    TitleString,
                    TitleString2,
                    TitleString3,
                    xyDataset,
                    PlotOrientation.VERTICAL,
                    displayLegend,
                    true,
                    false
            );
            final int minWidth = 300;
            final int minHeight = 200;
            final int width = 300;
            final int height = 200;
            final int maxWidth = 10000;
            final int maxHeight = 10000;
            chartsPanel.add(new ChartPanel(
                    chart,
                    width,
                    height,
                    minWidth,
                    minHeight,
                    maxWidth,
                    maxHeight,
                    false,
                    false,
                    true,
                    true,
                    true,
                    true
            ));
        }
        this.chartsPanel.validate();
    }

    public void fillSeries(
            final ArrayList<TrackSegment> tracks,
            final @NotNull Sequence sequence,
            final double stdGaussian,
            final double diskRadius,
            final double outterDiskRadius,
            final @NotNull AveragingType averagingMethod
    ) {
        // discriminate between new series and existing series to complement

        final ArrayList<TrackAnalysis> analyzedTrkList = new ArrayList<>();
        final boolean[] frameHasDetection = new boolean[sequence.getSizeT()];
        switch (averagingMethod) {
            case DISK: {
                final ArrayList<TrackAnalysisDiskAveraging> analysisList = new ArrayList<>();
                for (final TrackSegment ts : tracks) {
                    final TrackGroup group = ts.getOwnerTrackGroup();
                    final TrackAnalysisDiskAveraging analysis;
                    if (group == null)
                        analysis = new TrackAnalysisDiskAveraging(
                                ts,
                                sequence,
                                "Track " + trackPool.getTrackSegmentList().indexOf(ts),
                                averagingMethod
                        );
                    else
                        analysis = new TrackAnalysisDiskAveraging(
                                ts,
                                sequence,
                                group.getDescription() + "#" + group.getTrackSegmentList().indexOf(ts),
                                averagingMethod
                        );
                    analysisList.add(analysis);
                    analyzedTrkList.add(analysis);
                    final int t1 = Math.max(ts.getFirstDetection().getT(), 0);
                    final int t2 = Math.min(ts.getLastDetection().getT(), sequence.getSizeT() - 1);
                    for (int t = t1; t <= t2; t++)
                        frameHasDetection[t] = true;
                }
                for (int t = 0; t < sequence.getSizeT(); t++) {
                    if (frameHasDetection[t]) {
                        final double[][][] imageTab = new double[sequence.getSizeZ()][sequence.getSizeC()][];
                        for (int z = 0; z < sequence.getSizeZ(); z++)
                            for (int c = 0; c < sequence.getSizeC(); c++)
                                imageTab[z][c] = (double[]) ArrayUtil.arrayToDoubleArray(
                                        sequence.getImage(t, z, c).getDataXY(0),
                                        sequence.isSignedDataType()
                                );
                        for (final TrackAnalysisDiskAveraging ta : analysisList)
                            ta.fillAveragingSeriesAtT(t, imageTab, diskRadius);
                    }
                }
                break;
            }
            case SPOT_MASK: {
                final ArrayList<TrackAnalysisMaskAveraging> analysisList = new ArrayList<>();
                for (final TrackSegment ts : tracks) {
                    final TrackGroup group = ts.getOwnerTrackGroup();
                    final TrackAnalysisMaskAveraging analysis;
                    if (group == null)
                        analysis = new TrackAnalysisMaskAveraging(
                                ts,
                                sequence,
                                "Track " + trackPool.getTrackSegmentList().indexOf(ts),
                                averagingMethod
                        );
                    else
                        analysis = new TrackAnalysisMaskAveraging(
                                ts,
                                sequence,
                                group.getDescription() + "#" + group.getTrackSegmentList().indexOf(ts),
                                averagingMethod
                        );
                    analysisList.add(analysis);
                    analyzedTrkList.add(analysis);
                    final int minT = Math.min(ts.getLastDetection().getT(), sequence.getSizeT() - 1);
                    for (int t = ts.getFirstDetection().getT(); t <= minT; t++)
                        frameHasDetection[t] = true;
                }
                for (int t = 0; t < sequence.getSizeT(); t++) {
                    if (frameHasDetection[t]) {
                        final double[][][] imageTab = new double[sequence.getSizeZ()][sequence.getSizeC()][];
                        for (int z = 0; z < sequence.getSizeZ(); z++)
                            for (int c = 0; c < sequence.getSizeC(); c++)
                                imageTab[z][c] = (double[]) ArrayUtil.arrayToDoubleArray(
                                        sequence.getImage(t, z, c).getDataXY(0),
                                        sequence.isSignedDataType()
                                );
                        for (final TrackAnalysisMaskAveraging ta : analysisList)
                            ta.fillAveragingSeriesAtT(t, imageTab, diskRadius);
                    }
                }
                break;
            }
            case GAUSSIAN_FIT2D: {
                final ArrayList<TrackAnalysisGaussianFit> analysisList = new ArrayList<>();
                for (final TrackSegment ts : tracks) {
                    final TrackGroup group = ts.getOwnerTrackGroup();
                    final TrackAnalysisGaussianFit analysis;
                    if (group == null)
                        analysis = new TrackAnalysisGaussianFit(
                                ts,
                                sequence,
                                "Track " + trackPool.getTrackSegmentList().indexOf(ts),
                                averagingMethod
                        );
                    else
                        analysis = new TrackAnalysisGaussianFit(
                                ts,
                                sequence,
                                group.getDescription() + "#" + group.getTrackSegmentList().indexOf(ts),
                                averagingMethod
                        );
                    analysisList.add(analysis);
                    analyzedTrkList.add(analysis);
                    final int minT = Math.min(ts.getLastDetection().getT(), sequence.getSizeT() - 1);
                    for (int t = ts.getFirstDetection().getT(); t <= minT; t++)
                        frameHasDetection[t] = true;
                }
                for (int t = 0; t < sequence.getSizeT(); t++) {
                    if (frameHasDetection[t]) {
                        final double[][] imageTab = new double[sequence.getSizeC()][];
                        for (int c = 0; c < sequence.getSizeC(); c++)
                            imageTab[c] = (double[]) ArrayUtil.arrayToDoubleArray(
                                    sequence.getImage(t, 0, c).getDataXY(0),
                                    sequence.isSignedDataType()
                            );
                        for (final TrackAnalysisGaussianFit ta : analysisList)
                            ta.fillGaussianFitSeriesAtT(t, imageTab, stdGaussian);
                    }
                }
                break;
            }
            case DISK_BACKGROUND_CORRECTED: {
                final ArrayList<TrackAnalysisDiskAveragingNoBackground> analysisList = new ArrayList<>();
                for (final TrackSegment ts : tracks) {
                    final TrackGroup group = ts.getOwnerTrackGroup();
                    final TrackAnalysisDiskAveragingNoBackground analysis;
                    if (group == null)
                        analysis = new TrackAnalysisDiskAveragingNoBackground(
                                ts,
                                sequence,
                                "Track " + trackPool.getTrackSegmentList().indexOf(ts),
                                averagingMethod
                        );
                    else
                        analysis = new TrackAnalysisDiskAveragingNoBackground(
                                ts,
                                sequence,
                                group.getDescription() + "#" + group.getTrackSegmentList().indexOf(ts),
                                averagingMethod
                        );
                    analysisList.add(analysis);
                    analyzedTrkList.add(analysis);
                    for (int t = ts.getFirstDetection().getT(); t <= ts.getLastDetection().getT(); t++)
                        frameHasDetection[t] = true;
                }
                for (int t = 0; t < sequence.getSizeT(); t++) {
                    if (frameHasDetection[t]) {
                        final double[][][] imageTab = new double[sequence.getSizeZ()][sequence.getSizeC()][];
                        for (int z = 0; z < sequence.getSizeZ(); z++)
                            for (int c = 0; c < sequence.getSizeC(); c++)
                                imageTab[z][c] = (double[]) ArrayUtil.arrayToDoubleArray(
                                        sequence.getImage(t, z, c).getDataXY(0),
                                        sequence.isSignedDataType()
                                );
                        for (final TrackAnalysisDiskAveragingNoBackground ta : analysisList)
                            ta.fillAveragingSeriesAtT(t, imageTab, diskRadius, outterDiskRadius);
                    }
                }
                break;
            }
            default:
                break;
        }

        try {
            analyzedTracksLock.lock();
            analyzedTracks.addAll(analyzedTrkList);
            final ArrayList<TrackAnalysis> copyAnalyzedTracks = new ArrayList<>(analyzedTracks);
            SwingUtilities.invokeLater(() -> updateCharts(copyAnalyzedTracks, averagingMethod));
        }
        finally {
            analyzedTracksLock.unlock();
        }
    }
}
